/* Modélisation de la percolation de lien dans un réseau carré 2D
   Vincent MAGNIN, Polytech Lille (c)1990, 2000, 2003
   Version Java : 19 novembre 2003
   Dernière modification : 07/04/2013 adaptation à Java 7
   Compilation et exécution :
   javac percolation2D.java
   java percolation2D
*/

import java.awt.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.event.*;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
//import com.sun.image.codec.jpeg.*;    // plus disponible en java 7

public class percolation2D {
 static FenetrePrincipale miaFenestro;
 static Panneau pano; 
 static JProgressBar progression;
 static BufferedImage monImage;
 static WritableRaster tamponImage;
 static int LARGEUR=602;   //Taille de l'image en pixels
 static int HAUTEUR=602;

  //Variables globales de la modélisation :
  static boolean cepoint[];
  static int liens[][];

  static int pasx;
  static int pasy;
  static int pasx2;
  static int pasy2;

  static int nx;
  static int ny;

  static int nliens;

  static int xmax;
  static int ymax;

  static int max;

  // Couleurs disponibles :
  static double [] blanc = new double[]{255, 255, 255};
  static double [] noir = new double[]{0, 0, 0};
  static double [] bleu = new double[]{0, 0, 255};


 public static void main (String args[]) throws IOException {
  miaFenestro=new FenetrePrincipale ("Percolation 2D");
 }

 //Classe interne du panneau d'affichage :
  static class Panneau extends JPanel
 {
   //méthode appelée pour dessiner le contenu du panneau
   public void paintComponent (Graphics g)
   {
      //     super.paintComponent(g);
      g.drawImage(monImage, 0, 0, this);
    }
  }
  
  //Classe interne de la fenêtre :
 static class FenetrePrincipale extends JFrame implements KeyListener {
  static JTextField champTexte1;
    static JTextField champTexte2;
    static JTextField champTexte3;
    
  //Constructeur :
  FenetrePrincipale(String s) throws IOException {
       super(s); //Titre de la fenêtre

   //Si on veut fermer la fenêtre :
   addWindowListener(new WindowAdapter() {
    public void windowClosing(WindowEvent e) {
     System.exit(0);
    }
   });

   setBounds(0,0,1024,750); // Position et taille de la fenêtre

   GridBagLayout grilleConteneur= new GridBagLayout();
   GridBagConstraints contraintes= new GridBagConstraints();
   JPanel panneau = new JPanel();
   panneau.setLayout(grilleConteneur);

      //Champs pour entrer la température :
   construireContraintes(contraintes, 0, 0, 1, 1, 10, 10);
   contraintes.fill=GridBagConstraints.NONE;
   JLabel label1=new JLabel("Taux de liens restants :", JLabel.LEFT);
   grilleConteneur.setConstraints(label1, contraintes);
   panneau.add(label1);

   construireContraintes(contraintes, 1, 0, 1, 1, 90, 10);
   contraintes.fill=GridBagConstraints.NONE;
   champTexte1 = new JTextField("0.5",20);
   grilleConteneur.setConstraints(champTexte1, contraintes);
   panneau.add(champTexte1);

      //Champs pour entrer le nombre d'itérations :
   construireContraintes(contraintes, 0, 1, 1, 1, 10, 10);
   contraintes.fill=GridBagConstraints.NONE;
   JLabel label2=new JLabel("Largeur :", JLabel.LEFT);
   grilleConteneur.setConstraints(label2, contraintes);
   panneau.add(label2);

   construireContraintes(contraintes, 1, 1, 1, 1, 90, 10);
   contraintes.fill=GridBagConstraints.NONE;
   champTexte2 = new JTextField("300",20);
   grilleConteneur.setConstraints(champTexte2, contraintes);
   panneau.add(champTexte2);
   
   //Champs pour entrer le nombre de particules :
   construireContraintes(contraintes, 0, 2, 1, 1, 10, 10);
   contraintes.fill=GridBagConstraints.NONE;
   JLabel label3=new JLabel("Hauteur :", JLabel.LEFT);
   grilleConteneur.setConstraints(label3, contraintes);
   panneau.add(label3);

   construireContraintes(contraintes, 1, 2, 1, 1, 90, 10);
   contraintes.fill=GridBagConstraints.NONE;
   champTexte3 = new JTextField("300",20);
   grilleConteneur.setConstraints(champTexte3, contraintes);
   panneau.add(champTexte3);
 
      //Bouton  
   construireContraintes(contraintes, 0, 3, 1, 1, 0, 80);
   contraintes.fill=GridBagConstraints.NONE;
   JButton boutton1 = new JButton("Simuler !");
   grilleConteneur.setConstraints(boutton1, contraintes);
   panneau.add(boutton1);

      //Panneau d'affichage graphique :
   construireContraintes(contraintes, 1, 3, 1, 1, 90, 10);
   contraintes.fill=GridBagConstraints.NONE;
     pano= new Panneau(); //on déclare un panneau pour dessiner
     Dimension dim = new Dimension(LARGEUR,HAUTEUR);
     pano.setPreferredSize(dim);
   grilleConteneur.setConstraints(pano, contraintes);
   panneau.add(pano);   

      monImage=new BufferedImage(LARGEUR,HAUTEUR,BufferedImage.TYPE_INT_RGB);
      tamponImage=monImage.getRaster();
      
   //Si on clique sur le bouton :
   boutton1.addActionListener(new ActionListener() {
    public void actionPerformed(ActionEvent e) {
     try {
      modeliser();
     }
     catch (IOException ee) {
      System.out.println("Erreur d'entrée/sortie !");
     }
    }
   });

   //Si on appuie sur Entrée :
   champTexte1.addKeyListener(this);
   champTexte2.addKeyListener(this);
   champTexte3.addKeyListener(this);

      //Barre de progression :
   construireContraintes(contraintes, 0, 4, 1, 1, 0, 10);
   contraintes.fill=GridBagConstraints.NONE;
   progression= new JProgressBar(0, 100);
   progression.setValue(0);
   progression.setStringPainted(true);
   grilleConteneur.setConstraints(progression, contraintes);
   panneau.add(progression);
   
   setContentPane(panneau);
   
   //Création du menu :
   JMenuBar barreDeMenu = new JMenuBar();

   JMenu menu1 = new JMenu("Commandes", true);
   JMenuItem menu1_Item1 = new JMenuItem("Quitter");
   menu1_Item1.addActionListener(new ActionListener() {
    public void actionPerformed(ActionEvent e) {
     System.exit(0);
    }
   });
   menu1.add(menu1_Item1);

   JMenu menu2 = new JMenu("Aide", true);
   JMenuItem menu2_Item1 = new JMenuItem("A propos...");
   menu2_Item1.addActionListener(new ActionListener() {
    public void actionPerformed(ActionEvent e) {
     JOptionPane.showMessageDialog(null, "Percolation 2D (19-11-2003)\n");
    }
   });
   menu2.add(menu2_Item1);

   barreDeMenu.add(menu1);
   barreDeMenu.add(menu2);
   setJMenuBar(barreDeMenu);

   setVisible(true);
  }

  //**********************
  // Gestion du clavier :
  //**********************
  void construireContraintes(GridBagConstraints gbc, int gx, int gy, int gw, int gh, int wx, int wy) {
   gbc.gridx=gx;
   gbc.gridy=gy;
   gbc.gridwidth=gw;
   gbc.gridwidth=gh;
   gbc.weightx=wx;
   gbc.weighty=wy;
  }

  //**********************
  // Gestion du clavier :
  //**********************
  public void keyReleased(KeyEvent entree) {
   int touche = entree.getKeyCode();

   //Si on relache la touche Entrée :
   if (touche==10) {
    try {
     modeliser();
    }
    catch (IOException ee) {
     System.out.println("Erreur d'entrée/sortie !");
    }
   }

  }

  public void keyTyped(KeyEvent entree) {
   //Rien
  }

  public void keyPressed(KeyEvent entree) {
   //Rien
  }


    //*************************************************
 //*************************************************
 // Modélisation de la percolation de lien dans un réseau
 //*************************************************
    //*************************************************
    static void modeliser() throws IOException {
        int nb_liens; //Nombre de liens
        int i,j;
        double taux;
        
        // On efface l'image :
        for (i=0;i<=LARGEUR-1;i=i+1) {
            for (j=0;j<=HAUTEUR-1;j=j+1) {
            tamponImage.setPixel(i,j, blanc);
         }
        }
        pano.paintImmediately(0,0,LARGEUR,HAUTEUR);
        //Barre de progression à 0 :
        progression.setValue(0);
        progression.paintImmediately(0,0,progression.getWidth(),progression.getHeight());

        taux = Double.parseDouble(champTexte1.getText()); //Taux pas supprimé
        xmax = Integer.parseInt(champTexte2.getText()); //largeur du réseau
        ymax = Integer.parseInt(champTexte3.getText()); //hauteur

        nx = xmax + 1;
        ny = ymax + 1;
        max = ny * nx - 1;
        nliens = 4;        //réseau plan carré

        cepoint=new boolean[max+1];
        liens=new int[max+1][nliens+1];

        // Algorithme principal :
        initialise();
        nb_liens = compte_liens();
        //System.out.println("Nb initial de liens :"+nb_liens);
        elimine(nb_liens, taux);
        affichage();
        chemin();
        statistiques(taux);

        //Barre de progression :
        progression.setValue(100);
        progression.paintImmediately(0,0,progression.getWidth(),progression.getHeight());
        
        // On enregistre l'image au format JPEG :
//        FileOutputStream sortie=new FileOutputStream(new File("barreau2D.jpg"));
//        JPEGImageEncoder monEncodeur=JPEGCodec.createJPEGEncoder(sortie);
//        JPEGEncodeParam param=monEncodeur.getDefaultJPEGEncodeParam(monImage);
//        param.setQuality(0.5f, false);    //Sans compression
//        monEncodeur.setJPEGEncodeParam(param);
//        monEncodeur.encode(monImage);
//        sortie.close();
    }
    
    static void affichage () {
      int x;
      int y;
      int xx;
      int yy;
      int i,j;
      int p;

      pasy = HAUTEUR / ny;
      pasx = pasy;
      pasx2 = pasx / 2;
      pasy2 = pasy / 2;

      for (i = 0; i<=max; i++) {
        y = i / nx;    //division entière
        x = (i - y * nx) * pasx + pasx2;
        y = y * pasy + pasy2;

        tamponImage.setPixel(x,y, noir);

        for (j = 2; j<=3; j++) {
          p = liens[i][j];
          if (p >= 0) {
            yy = p / nx;      //division entière
            xx = (p - yy * nx) * pasx + pasx2;
            yy = yy * pasy + pasy2;
            traceLigne(x, y, xx, yy, noir);
          }
        }
      }
      pano.paintImmediately(0,0,LARGEUR,HAUTEUR);
    }

    static void traceLigne(int x, int y, int xx, int yy, double[] couleur)  {
        int i;
        double j,dy;
        
        if ((xx-x)!=0) {
          dy=(yy-y)/(1.0*(xx-x));
          j=y;
          for (i=x;i<=xx;i++) {
              j=j+dy;
              tamponImage.setPixel(i,(int)j, couleur);
          }
        }
        else
        {
            for (i=y;i<=yy;i++) {
                tamponImage.setPixel(x, i, couleur);
            }
        }
    }
    
    static void traceRectangle(int x, int y, int xx, int yy, double[] couleur)  {
        int i,j;

        for (i=x;i<=xx;i++) {
            for (j=y;j<=yy;j++) {
                tamponImage.setPixel(i,j, couleur);
          }
        }
    }
    
    static void chemin() {
      int pile[]=new int[4000];
      int k ;
      int i ;
      int der ;
      int top ;
      int x ;
      int y ;
      int n ;
      int j ;
      int p ;

      for (k = max - xmax; k<=max; k++) {
        pile[1] = k;
        der = 1;
        top = 2;
        y = k / nx;       //Division entière
        x = (k - y * nx) * pasx + pasx2;
        y = y * pasy + pasy2;
        traceRectangle(x-1,y-1,x,y,bleu);
        cepoint[k] = true;
        do  {
          for (i = 1; i<=der; i++) {
            n = pile[i];
            for (j = 1; j<=nliens; j++) {
              // if (j == 2) {j = 3;}   //percolation dirigée (retour interdit)
              p = liens[n][j];
              if (p!=-1) {
                if (!cepoint[p]) {
                  pile[top] = p;
                  cepoint[p] = true;
                  top = top + 1;
                  y = p / nx;  //Division entière
                  x = (p - y * nx) * pasx + pasx2;
                  y = y * pasy + pasy2;
                  traceRectangle(x-1,y-1,x,y,bleu);
                }
              }
            }
          }
          for (j = der + 1; j<=top; j++) {
            pile[j - der] = pile[j];
          }
          top = top - der;
          der = top - 1;
        } while (der != 0);
      }
      pano.paintImmediately(0,0,LARGEUR,HAUTEUR);
    }

    // Compte le nombre total de liens dans le réseau :
    //
    static int compte_liens () {
      int nb_total_liens;
      int i;
      int j;

      nb_total_liens = 0;

      for (i = 0; i<=max; i=i+1) {
        for (j = 1 ; j<=nliens ; j=j+1) {
          if (liens[i][j]!=-1) {
            nb_total_liens = nb_total_liens + 1;
          }
        }
      }

      return nb_total_liens/2;
    }

    //  On élimine une certaine proportion de liens :
    static void elimine(int nb_total_liens, double taux) {
      int n;
      int p;
      int i;
      int j;
      int q;

      i = nb_total_liens;

      do {
        n = 1 + (int)(Math.random()*nliens);
        p = (int)(Math.random()*(1 + max));

        q = liens[p][n];
        if (q!=-1) {
          liens[p][n] = -1;
          for (j=1; j<=nliens; j=j+1) {
            if (liens[q][j]==p) {
              liens[q][j] = -1;
            }
          }
          i = i - 1;
        }
      } while (i >(nb_total_liens * taux));

    }

    static void initialise () {
      int n,x,y;

      for (x = 1; x<=xmax - 1;x=x+1) {
        for (y = 1; y<=ymax - 1; y=y+1) {
          n = x + y * nx;
          //System.out.println(x+" "+y+" "+nx+" "+n);
          liens[n][1] = x + (y - 1) * nx;    //dessus
          liens[n][2] = x + (y + 1) * nx;    //dessous
          liens[n][4] = (x - 1) + y * nx;    //gauche
          liens[n][3] = (x + 1) + y * nx;    //droite
        }
      }
      x = 0;
      for (y = 1; y<=ymax - 1; y=y+1) {
          n = x + y * nx;
          liens[n][1] = x + (y - 1) * nx;    //dessus
          liens[n][2] = x + (y + 1) * nx;    //dessous
          liens[n][4] = -1;                  //gauche
          liens[n][3] = (x + 1) + y * nx;    //droite
      }
      x = xmax;
      for (y = 1; y<=ymax - 1; y=y+1) {
          n = x + y * nx;
          liens[n][1] = x + (y - 1) * nx;    //dessus
          liens[n][2] = x + (y + 1) * nx;    //dessous
          liens[n][4] = (x - 1) + y * nx;    //gauche
          liens[n][3] = -1;                  //droite
      }
      y = 0;
      for (x = 1; x<=xmax - 1;x=x+1) {
          n = x + y * nx;
          liens[n][1] = -1;                  //dessus
          liens[n][2] = x + (y + 1) * nx;    //dessous
          liens[n][4] = (x - 1) + y * nx;    //gauche
          liens[n][3] = (x + 1) + y * nx;    //droite
      }
      y = ymax;
      for (x = 1; x<=xmax - 1;x=x+1) {
          n = x + y * nx;
          liens[n][1] = x + (y - 1) * nx;    //dessus
          liens[n][2] = -1;                  //dessous
          liens[n][4] = (x - 1) + y * nx;    //gauche
          liens[n][3] = (x + 1) + y * nx;    //droite
      }
      x = 0;
      y = 0;
      n = x + y * nx;
      liens[n][1] = -1;                  //dessus
      liens[n][2] = x + (y + 1) * nx;    //dessous
      liens[n][4] = -1;                  //gauche
      liens[n][3] = (x + 1) + y * nx;    //droite
      y = ymax;
      n = x + y * nx;
      liens[n][1] = x + (y - 1) * nx;    //dessus
      liens[n][2] = -1;                 //dessous
      liens[n][4] = -1;                 //gauche
      liens[n][3] = (x + 1) + y * nx;    //droite
      x = xmax;
      y = 0;
      n = x + y * nx;
      liens[n][1] = -1;              //dessus
      liens[n][2] = x + (y + 1) * nx;    //dessous
      liens[n][4] = (x - 1) + y * nx;    //gauche
      liens[n][3] = -1;              //droite
      y = ymax;
      n = x + y * nx;
      liens[n][1] = x + (y - 1) * nx;    //dessus
      liens[n][2] = -1;      //dessous
      liens[n][4] = (x - 1) + y * nx;    //gauche
      liens[n][3] = -1;      //droite

      for (n = 0; n<=max; n=n+1) {
        cepoint[n] = false;
      }
    }


    static void statistiques (double taux) {
      int i;
      double atteint;
      double passe;

      System.out.print("p= "+taux+" ");
            
      atteint = 0;
      for (i = 0; i<=max; i++) {
        if (cepoint[i]) {
          atteint = atteint + 1;
        }
      }
      System.out.print(", points atteints r= " + (Math.round(atteint / (max + 1) * 1000) / 10.0) + " % ");

      passe = 0;
      for (i = 0; i<=xmax; i++) {
        if (cepoint[i]) {
          passe = passe + 1;
        }
      }
      System.out.println(", points en haut q= " + (Math.round(passe / (xmax + 1) * 1000) / 10.0) + "%");
    }

  } // Fin de la classe interne "FenetrePrincipale"

} // Fin de la classe barreau2D
